! WHIZARD 2.2.0_alpha-1 Sep 25 2013
! 
! Copyright (C) 1999-2013 by 
!     Wolfgang Kilian <kilian@physik.uni-siegen.de>
!     Thorsten Ohl <ohl@physik.uni-wuerzburg.de>
!     Juergen Reuter <juergen.reuter@desy.de>
!     
!     with contributions by 
!     Christian Speckner <cnspeckn@googlemail.com> 
!     and  Felix Braam, Sebastian Schmidt, Daniel Wiesler 
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by 
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

module integrations

  use kinds, only: default !NODEP!
  use iso_varying_string, string_t => varying_string !NODEP!
  use file_utils !NODEP!
!  use limits, only: ITERATIONS_DEFAULT_LIST_SIZE !NODEP!
  use diagnostics !NODEP!
  use unit_tests
!  use tao_random_numbers !NODEP!
!  use pdf_builtin !NODEP!
  use os_interface
  use sm_qcd
  use ifiles
  use lexers
  use parser
  use flavors
  use variables
  use expressions
  use models
  use interactions
  use beams
  use sf_mappings
  use sf_base
  use phs_base
  use phs_forests
  use phs_wood
  use rng_base
  use mci_base
  use process_libraries
  use prc_core
  use processes
  use process_stacks
!  use strfun_config
  use rt_data
  use dispatch
!  use iterations
  use process_configurations
  use compilations

  implicit none
  private

  public :: integrate_process
  public :: integrations_test

  type :: integration_t
    private
    type(string_t) :: process_id
    type(string_t) :: run_id
    type(process_t), pointer :: process => null ()
    logical :: rebuild_phs = .false.
!     logical :: check_phs_file = .true.
!     type(string_t) :: phs_filename
!     type(string_t) :: phs_filename_out
!     type(string_t) :: phs_filename_vis
    logical :: phs_only = .false.
!     logical :: vis_channels = .false.
!     type(phs_parameters_t) :: phs_par
!     type(mapping_defaults_t) :: mapping_defaults
!     logical :: rebuild_grids = .false.
!     logical :: check_grid_file = .true.
!     type(grid_parameters_t) :: grid_parameters
!     type(string_t) :: grids_filename
!     logical :: use_best_grid = .true.
!     real(default) :: accuracy_goal = 0
!     real(default) :: abs_error_goal = 0
!     real(default) :: rel_error_goal = 0
!     logical :: helicity_selection_active = .false.
!     real(default) :: helicity_selection_threshold = -1
!     integer :: helicity_selection_cutoff = 1000
!     logical :: sqrts_known = .false.
!     real(default) :: sqrts = -1
!     real(default) :: alpha_s = -1
!     real(default) :: alpha = 0
!     integer :: n_events_for_me_test = 0
!     logical :: time_estimate = .false.
!     logical :: vis_history = .true.
!     type(string_t) :: history_filename
!     type(beam_data_t) :: beam_data
!     logical :: use_beams = .false.
!     type(sf_list_t), pointer :: sf_list => null ()
!     logical :: use_strfun = .false.
!     type(iterations_list_t) :: it_list
!     integer :: pass_on_file = 0
!     integer :: it_on_file = 0
!     type(md5sum_grids_t) :: md5sum
!     integer :: pass = 0
!     integer :: it = 0
!     type(string_t) :: log_filename
   contains
     procedure :: create_process => integration_create_process
     procedure :: setup_process => integration_setup_process
     procedure :: evaluate => integration_evaluate
     generic :: init => integration_init0 !, integration_init1
     procedure :: integration_init0
   !  procedure :: integration_init1
     generic :: integrate => integration_integrate0 !, integration_integrate1
     procedure :: integration_integrate0
   !  procedure :: integration_integrate1 
  end type integration_t


  interface integrate_process
     module procedure integrate_process0
!     module procedure integrate_process1
  end interface


contains

  subroutine integration_create_process (intg, process_id, global) !, verbose)
    class(integration_t), intent(out) :: intg
    type(rt_data_t), target :: global
    type(string_t), intent(in) :: process_id
!    logical, intent(in), optional :: verbose
    type(var_list_t), pointer :: var_list
!     logical :: verb
    type(process_entry_t), pointer :: process_entry

    var_list => global%var_list
    intg%process_id = process_id
    intg%run_id = var_list_get_sval (var_list, var_str ("$run_id"))

    allocate (process_entry)
    call process_entry%init (intg%process_id, intg%run_id, global%prclib, &
         global%os_data)
    call process_entry%set_var_list (global%var_list)
    call global%process_stack%push (process_entry)

  end subroutine integration_create_process

  subroutine integration_setup_process (intg, global)
    class(integration_t), intent(inout) :: intg
    type(rt_data_t), intent(inout), target :: global

    class(prc_core_t), allocatable :: core_template
    class(phs_config_t), allocatable :: phs_config_template
    class(mci_t), allocatable :: mci_template
    type(qcd_t) :: qcd
    integer :: n_components, i_component
    type(process_component_def_t), pointer :: config
    type(helicity_selection_t), allocatable :: helicity_selection
    real(default) :: sqrts
    type(sf_config_t), dimension(:), allocatable :: sf_config
    type(sf_channel_t), dimension(:), allocatable :: sf_channel
    class(rng_t), allocatable :: rng
    logical :: rng_seed_known
    integer :: rng_seed
    integer :: n_mci, i_mci

    intg%process => global%process_stack%get_process_ptr (intg%process_id)

!    intg%rng_method = &
!         var_list_get_sval (var_list, var_str ("$rng_method"))

    intg%rebuild_phs = &
         var_list_get_lval (global%var_list, var_str ("?rebuild_phase_space"))
!     intg%check_phs_file = &
!          var_list_get_lval (var_list, var_str ("?check_phs_file"))
!     intg%phs_filename = &
!          var_list_get_sval (var_list, var_str ("$phs_file"))
    intg%phs_only = &
         var_list_get_lval (global%var_list, var_str ("?phs_only"))
!     intg%vis_channels = &
!          var_list_get_lval (var_list, var_str ("?vis_channels"))
!     intg%phs_par%m_threshold_s = &
!          var_list_get_rval (var_list, var_str ("phs_threshold_s"))
!     intg%phs_par%m_threshold_t = &
!          var_list_get_rval (var_list, var_str ("phs_threshold_t"))
!     intg%phs_par%off_shell = &
!          var_list_get_ival (var_list, var_str ("phs_off_shell"))
!     intg%phs_par%keep_nonresonant = &
!          var_list_get_lval (var_list, var_str ("?phs_keep_nonresonant"))
!     intg%phs_par%t_channel = &
!          var_list_get_ival (var_list, var_str ("phs_t_channel"))
!     intg%mapping_defaults%energy_scale = &
!          var_list_get_rval (var_list, var_str ("phs_e_scale"))
!     intg%mapping_defaults%invariant_mass_scale = &
!          var_list_get_rval (var_list, var_str ("phs_m_scale"))
!     intg%mapping_defaults%momentum_transfer_scale = &
!          var_list_get_rval (var_list, var_str ("phs_q_scale"))
!     intg%mapping_defaults%step_mapping = &
!          var_list_get_lval (var_list, var_str ("?phs_step_mapping"))
!     intg%mapping_defaults%step_mapping_exp = &
!          var_list_get_lval (var_list, var_str ("?phs_step_mapping_exp"))
!     intg%mapping_defaults%enable_s_mapping = &
!          var_list_get_lval (var_list, var_str ("?phs_s_mapping"))

    call dispatch_phs (phs_config_template, global)
    
!     intg%rebuild_grids = &
!          var_list_get_lval (var_list, var_str ("?rebuild_grids"))
!     intg%check_grid_file = &
!          var_list_get_lval (var_list, var_str ("?check_grid_file"))

!     intg%run_id = &
!          var_list_get_sval (var_list, var_str ("$run_id"))  ! $
!     if (intg%run_id /= "") then
!        intg%phs_filename_out = intg%process_id // "." // intg%run_id // ".phs"
!        intg%phs_filename_vis = intg%process_id // "." // intg%run_id // "_phs"
!        intg%grids_filename = intg%process_id // "." // intg%run_id // ".vg"
!        intg%history_filename = intg%process_id // "." // intg%run_id &
!             // "-history"
!        intg%log_filename = intg%process_id // "." // intg%run_id // ".log"
!     else
!        intg%phs_filename_out = intg%process_id // ".phs"
!        intg%phs_filename_vis = intg%process_id // "_phs"
!        intg%grids_filename = intg%process_id // ".vg"
!        intg%history_filename = intg%process_id // "-history"
!        intg%log_filename = intg%process_id // ".log"
!     end if
!     intg%use_best_grid = &
!          var_list_get_lval (var_list, var_str ("?use_best_grid"))
!     intg%accuracy_goal = &
!          var_list_get_rval (var_list, var_str ("accuracy_goal"))
!     intg%abs_error_goal = &
!          var_list_get_rval (var_list, var_str ("error_goal"))
!     intg%rel_error_goal = &
!          var_list_get_rval (var_list, var_str ("relative_error_goal"))

    call dispatch_mci (mci_template, global, intg%process_id)

!     verb = .true.;  if (present (verbose))  verb = verbose
!     if (verb) then
!        call msg_message ("Initializing integration for process " &
!             // char (intg%process_id) // ":")
!        if (intg%run_id /= "") then
!           call msg_message ("Run ID = " // '"' // char (intg%run_id) // '"')
!        end if
!     end if
    
    helicity_selection = global%get_helicity_selection ()

!     intg%sqrts_known = var_list_is_known (var_list, "sqrts")
!     intg%sqrts = var_list_get_rval (var_list, "sqrts")
!     intg%time_estimate = &
!          var_list_get_lval (var_list, var_str ("?time_estimate"))
!     intg%vis_history = &
!          var_list_get_lval (var_list, var_str ("?vis_history"))
!     intg%n_events_for_me_test = &
!          var_list_get_ival (var_list, var_str ("n_events"))
!     if (associated (global%pn_alpha_qed_expr)) then
!        intg%alpha = eval_real (global%pn_alpha_qed_expr, var_list)
!     else
!        intg%alpha = 0
!     end if
! 

    call dispatch_qcd (qcd, global)
    
    n_components = intg%process%get_n_components ()
    do i_component = 1, n_components
       config => intg%process%get_component_def_ptr (i_component)
       call dispatch_core (core_template, config%get_core_def_ptr (), &
            intg%process%get_model_ptr (), helicity_selection, qcd)
       call intg%process%init_component &
            (i_component, core_template, mci_template, phs_config_template)
    end do
    
    if (beam_data_are_valid (global%beam_data)) then
       call intg%process%setup_beams (global%beam_data)
    else if (intg%process%get_n_in () == 2) then
       sqrts = var_list_get_rval (global%var_list, var_str ("sqrts"))
       call intg%process%setup_beams (sqrts)
    else 
       call msg_bug ("Decay process: Automatic beam setup not supported yet")
!       call intg%process%setup_beams ()
    end if
    call intg%process%beams_startup_message ()

    call dispatch_sf_config (sf_config, sf_channel, global)
    if (size (sf_config) > 0) then
       call intg%process%init_sf_chain (sf_config)
    end if

    call intg%process%configure_phs (intg%rebuild_phs)

    if (allocated (sf_channel)) then
       if (size (sf_channel) > 0) then
          call intg%process%set_sf_channel (sf_channel)
       end if
    end if
    call intg%process%match_channels ()
    call intg%process%sf_startup_message ()
    
    call intg%process%setup_mci ()

    rng_seed_known = var_list_is_known (global%var_list, var_str ("seed"))
    if (rng_seed_known) then
       rng_seed = var_list_get_ival (global%var_list, var_str ("seed"))
    end if
    n_mci = intg%process%get_n_mci ()
    do i_mci = 1, n_mci
       call dispatch_rng (rng, global)
       if (rng_seed_known) then
          call rng%init (rng_seed + i_mci - 1)
       else
          call rng%init ()
       end if
       call intg%process%import_rng (i_mci, rng)
    end do
       
    call intg%process%setup_terms ()
    
    call intg%process%set_cuts (global%pn%cuts_lexpr)
    call intg%process%set_scale (global%pn%scale_expr)
    call intg%process%set_fac_scale (global%pn%fac_scale_expr)
    call intg%process%set_ren_scale (global%pn%ren_scale_expr)
    call intg%process%set_weight (global%pn%weight_expr)

!     call process_set_alpha_qed (intg%process, intg%alpha)
!     call process_reset_helicity_selection (intg%process, &
!          intg%helicity_selection_threshold, intg%helicity_selection_cutoff)

    call intg%process%compute_md5sum ()
    
  end subroutine integration_setup_process

  subroutine integration_evaluate (intg, process_instance, i_mci, pass, global)
!        (intg, rng, pass, global_var_list, os_data, verbose)
    class(integration_t), intent(inout) :: intg
    type(process_instance_t), intent(inout), target :: process_instance
!     type(tao_random_state), intent(inout) :: rng
    integer, intent(in) :: i_mci
    integer, intent(in) :: pass
    type(rt_data_t), intent(in), target :: global
!     type(var_list_t), intent(inout) :: global_var_list
!     type(os_data_t), intent(in) :: os_data
!     logical, intent(in), optional :: verbose
!     integer :: it_on_file, u
    integer :: n_calls, n_it
    logical :: adapt_grids, adapt_weights
!     logical :: verb
!     verb = .true.;  if (present (verbose))  verb = verbose
!     u = logfile_unit ()
!     intg%pass = pass
!     if (intg%pass == intg%pass_on_file) then
!        it_on_file = intg%it_on_file
!     else
!        it_on_file = 0
!     end if

    associate (it_list => global%it_list)
      if (pass <= it_list%get_n_pass ()) then
         n_calls = it_list%get_n_calls (pass)
         n_it    = it_list%get_n_it (pass)
         adapt_grids   = it_list%adapt_grids (pass)
         adapt_weights = it_list%adapt_weights (pass)
      else
         n_calls = intg%process%get_n_calls_default (pass)
         n_it    = intg%process%get_n_it_default (pass)
         adapt_grids   = intg%process%adapt_grids_default (pass)
         adapt_weights = intg%process%adapt_weights_default (pass)
      end if
    end associate
    !     if (iterations_list_has_custom_adaptation (intg%it_list, intg%pass)) then
!        adapt_grids = iterations_list_adapt_grids (intg%it_list, intg%pass)
!        adapt_weights = iterations_list_adapt_weights (intg%it_list, intg%pass)
!     else
!        adapt_grids = .true.
!        adapt_weights = .false.
!     end if
!     do i = 1, it_on_file
!        intg%it = intg%it + 1
!        if (verb) then
!           call process_results_write_entry (intg%process, intg%it)
!           if (u > 0) then
!              call process_results_write_entry (intg%process, intg%it, unit=u)
!              flush (u)
!           end if
!        end if
!     end do

    call intg%process%integrate (process_instance, &
         i_mci, n_it, n_calls, adapt_grids, adapt_weights)

!      call process_integrate (intg%process, rng, &
!          intg%grid_parameters, &
!          intg%pass, it_on_file + 1, n_it, n_calls, &
!          discard_integrals = .true., &
!          adapt_grids = adapt_grids, &
!          adapt_weights = adapt_weights, &
!          print_current = verb, &
!          time_estimate = intg%time_estimate, &
!          grids_filename = intg%grids_filename, &
!          write_best_grid = intg%use_best_grid, &
!          md5sum = intg%md5sum, &
!          history_filename = intg%history_filename, &
!          log_filename = intg%log_filename)
!     if (verb) then
!        call process_results_write_average (intg%process, intg%pass)
!        if (u > 0) then
!           call process_results_write_average (intg%process, intg%pass, unit=u)
!           flush (u)
!        end if
!     end if
!     if (intg%vis_history) then
!        call process_display_integration_history &
!             (intg%process, intg%history_filename, os_data)
!     end if
!     call process_record_integral (intg%process, global_var_list)

  end subroutine integration_evaluate

  subroutine integration_init0 (intg, process_id, global)
!      (intg, process_id, global, ok, me_only, no_beams, verbose)
    class(integration_t), intent(out) :: intg
    type(string_t), intent(in) :: process_id
    type(rt_data_t), intent(inout), target :: global
!     logical, intent(out) :: ok
!     logical, intent(in), optional :: me_only, no_beams, verbose
!     logical :: integrate, allow_beams
!     integrate = .true.;  if (present (me_only))  integrate = .not. me_only
!     allow_beams = .true.;  if (present (no_beams)) allow_beams = .not. no_beams
    call intg%create_process (process_id, global) !, verbose)

!     if (allow_beams)  call integration_check_beam_data (intg, global%beam_data)
!     call maybe_compile_library (global)

    call intg%setup_process (global)

!     if (ok) then
!        call integration_setup_beams &
!             (intg, global%sf_list, global%var_list, ok)
!     end if
!     if (ok) then
!        call integration_setup_qcd &
!             (intg, global%lhapdf_status, global%pdf_builtin_status, &
!              global%os_data, global%var_list)
!     end if
!     if (integrate .and. ok) then
!        call integration_setup_phase_space (intg, global%os_data, ok)
!     end if
!     if (integrate .and. ok) then
!        call integration_setup_strfun_mappings (intg)
!     end if
!     if (.not. intg%phs_only) then
!        if (integrate .and. ok) then
!           call integration_collect_md5sums (intg)
!           call integration_setup_iterations &
!                (intg, global%it_list, global%it_list_default, ok, verbose)
!        end if
!        if (ok) then
!           call integration_setup_subevt (intg)
!           call integration_setup_cuts (intg, global%pn_cuts_lexpr, verbose)
!           call integration_setup_scale (intg, global%pn_scale_expr, verbose)    
!           call integration_setup_fac_scale (intg, global%pn_fac_scale_expr, verbose)
!           call integration_setup_ren_scale (intg, global%pn_ren_scale_expr, verbose)    
!           call integration_setup_weight (intg, global%pn_weight_expr, verbose)
!        end if
!        if (integrate .and. ok) then
!           call integration_setup_grids (intg)
!        end if
!     end if
  end subroutine integration_init0

  subroutine integration_integrate0 (intg, global)
!       (intg, rng, global_var_list, os_data, verbose)
    class(integration_t), intent(inout) :: intg
    type(rt_data_t), intent(inout), target :: global
!    type(tao_random_state), intent(inout) :: rng
!    type(var_list_t), intent(inout) :: global_var_list
!    type(os_data_t), intent(in) :: os_data
!    logical, intent(in), optional :: verbose

    type(process_instance_t), allocatable, target :: process_instance
    integer :: pass, i_mci, n_mci, n_pass

!     call openmp_set_num_threads_verbose &
!        (var_list_get_ival (global_var_list, var_str ("openmp_num_threads")))
!     call integration_write_header (intg, verbose)

    allocate (process_instance)
    call process_instance%init (intg%process)

!     do pass = 1, iterations_list_get_n_pass (intg%it_list) - 1
!        call integration_warmup (intg, rng, pass, verbose)
!     end do
!     if (intg%use_best_grid) &
!          call process_choose_best_grid (intg%process, intg%check_grid_file)

    n_mci = intg%process%get_n_mci ()
    if (n_mci == 1) then
       write (msg_buffer, "(A,A,A)") &
            "Starting integration for process '", &
            char (intg%process%get_id ()), "'"
       call msg_message ()
    end if
    do i_mci = 1, n_mci
       if (n_mci > 1) then
          write (msg_buffer, "(A,A,A,I0)") &
               "Starting integration for process '", &
               char (intg%process%get_id ()), "' part ", i_mci
          call msg_message ()
       end if
       n_pass = global%it_list%get_n_pass ()
       if (n_pass == 0) then
          call msg_message ("Integrate: iterations not specified, &
               &using default")
          n_pass = intg%process%get_n_pass_default ()
       end if
       do pass = 1, n_pass
          call intg%evaluate (process_instance, i_mci, pass, global)
       end do
       call intg%process%final_integration (i_mci)
    end do
       
!         (intg, rng, pass, global_var_list, os_data, verbose)

!     call integration_write_footer (intg, verbose)

    call process_instance%final ()
    deallocate (process_instance)

  end subroutine integration_integrate0

  subroutine integrate_process0 (process_id, global)
!      (process_id, global, global_var_list, no_beams, verbose)
    type(string_t), intent(in) :: process_id
    type(rt_data_t), intent(inout), target :: global
!     type(var_list_t), intent(inout) :: global_var_list
!     logical, intent(in), optional :: no_beams, verbose
    type(string_t) :: prclib_name
    type(integration_t) :: intg
!     logical :: ok

    if (.not. associated (global%prclib)) then
       call msg_fatal ("Integrate: current process library is undefined")
       return
    end if

    if (.not. global%prclib%is_active ()) then
       call msg_message ("Integrate: current process library needs compilation")
       prclib_name = global%prclib%get_name ()
       call compile_library (prclib_name, global)
       call msg_message ("Integrate: compilation done")
    end if

    call intg%init (process_id, global)
!         (intg, process_id, global, ok, no_beams=no_beams, verbose=verbose)

    if (intg%phs_only) then
       call msg_message ("Integrate: phase space only, skipping integration")
    else
       call intg%integrate (global)
    end if


!     if (.not. intg%phs_only) then
!        if (ok) then
!           call integration_integrate &
!                (intg, global%rng, global_var_list, global%os_data, verbose)
!        else
!           call integration_integrate_dummy (intg, global_var_list, verbose)
!        end if
!     end if
  end subroutine integrate_process0

!   subroutine integrate_process1 &
!        (process_id, global, global_var_list, no_beams, verbose)
!     type(string_t), dimension(:), intent(in) :: process_id
!     type(rt_data_t), intent(inout), target :: global
!     type(var_list_t), intent(inout) :: global_var_list
!     logical, intent(in), optional :: no_beams, verbose
!     integer :: proc
!     do proc = 1, size (process_id)
!        call integrate_process0 &
!             (process_id(proc), global, global_var_list, no_beams, verbose)
!     end do
!   end subroutine integrate_process1
! 

  subroutine integrations_test (u, results)
    integer, intent(in) :: u
    type(test_results_t), intent(inout) :: results
    call test (integrations_1, "integrations_1", &
         "intrinsic test process", &
         u, results)
    call test (integrations_2, "integrations_2", &
         "intrinsic test process with cut", &
         u, results)
    call test (integrations_3, "integrations_3", &
         "standard phase space", &
         u, results)
    call test (integrations_4, "integrations_4", &
         "VAMP integration (one iteration)", &
         u, results)
    call test (integrations_5, "integrations_5", &
         "VAMP integration (three iterations)", &
         u, results)
    call test (integrations_6, "integrations_6", &
         "VAMP integration (three passes)", &
         u, results)
    call test (integrations_7, "integrations_7", &
         "VAMP integration with wood phase space", &
         u, results)
    call test (integrations_8, "integrations_8", &
         "integration with structure function", &
         u, results)
  end subroutine integrations_test

  subroutine integrations_1 (u)
    integer, intent(in) :: u
    type(string_t) :: libname, procname
    type(rt_data_t), target :: global
    
    write (u, "(A)")  "* Test output: integrations_1"
    write (u, "(A)")  "*   Purpose: integrate test process"
    write (u, "(A)")

    call syntax_model_file_init ()

    call global%global_init ()

    libname = "integration_1"
    procname = "prc_config_a"
    
    call prepare_test_library (global, libname, 1)
    call compile_library (libname, global)

    call var_list_set_string (global%var_list, var_str ("$run_id"), &
         var_str ("integrations1"), is_known = .true.)
    call var_list_set_string (global%var_list, var_str ("$method"), &
         var_str ("unit_test"), is_known = .true.)
    call var_list_set_string (global%var_list, var_str ("$phs_method"), &
         var_str ("single"), is_known = .true.)
    call var_list_set_string (global%var_list, var_str ("$integration_method"),&
         var_str ("midpoint"), is_known = .true.)

    call var_list_set_real (global%var_list, var_str ("sqrts"),&
         1000._default, is_known = .true.)

    call global%it_list%init ([1], [1000])

    call reset_interaction_counter ()
    call integrate_process (procname, global)

    call global%write (u, vars = [ &
         var_str ("$method"), &
         var_str ("sqrts"), &
         var_str ("$integration_method"), &
         var_str ("$phs_method"), &
         var_str ("$run_id")])
    
    call global%final ()
    call model_list_final ()
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: integrations_1"
    
  end subroutine integrations_1
  
  subroutine integrations_2 (u)
    integer, intent(in) :: u
    type(string_t) :: libname, procname
    type(rt_data_t), target :: global

    type(string_t) :: cut_expr_text
    type(ifile_t) :: ifile
    type(stream_t) :: stream
    type(parse_tree_t) :: parse_tree
    
    type(string_t), dimension(0) :: empty_string_array

    write (u, "(A)")  "* Test output: integrations_2"
    write (u, "(A)")  "*   Purpose: integrate test process with cut"
    write (u, "(A)")

    call syntax_model_file_init ()

    call global%global_init ()

    write (u, "(A)")  "* Prepare a cut expression"
    write (u, "(A)")

    call syntax_pexpr_init ()
    cut_expr_text = "all Pt > 100 [s]"
    call ifile_append (ifile, cut_expr_text)
    call stream_init (stream, ifile)
    call parse_tree_init_lexpr (parse_tree, stream, .true.)
    global%pn%cuts_lexpr => parse_tree_get_root_ptr (parse_tree)
    
    write (u, "(A)")  "* Build and initialize a test process"
    write (u, "(A)")

    libname = "integration_3"
    procname = "prc_config_a"
    
    call prepare_test_library (global, libname, 1)
    call compile_library (libname, global)

    call var_list_set_string (global%var_list, var_str ("$run_id"), &
         var_str ("integrations1"), is_known = .true.)
    call var_list_set_string (global%var_list, var_str ("$method"), &
         var_str ("unit_test"), is_known = .true.)
    call var_list_set_string (global%var_list, var_str ("$phs_method"), &
         var_str ("single"), is_known = .true.)
    call var_list_set_string (global%var_list, var_str ("$integration_method"),&
         var_str ("midpoint"), is_known = .true.)

    call var_list_set_real (global%var_list, var_str ("sqrts"),&
         1000._default, is_known = .true.)

    call global%it_list%init ([1], [1000])
    
    call reset_interaction_counter ()
    call integrate_process (procname, global)
    
    call global%write (u, vars = empty_string_array)
    
    call global%final ()
    call model_list_final ()
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: integrations_2"
    
  end subroutine integrations_2
  
  subroutine integrations_3 (u)
    integer, intent(in) :: u
    type(string_t) :: libname, procname
    type(rt_data_t), target :: global
    integer :: u_phs
    
    write (u, "(A)")  "* Test output: integrations_3"
    write (u, "(A)")  "*   Purpose: integrate test process"
    write (u, "(A)")

    write (u, "(A)")  "* Initialize process and parameters"
    write (u, "(A)")

    call syntax_model_file_init ()
    call syntax_phs_forest_init ()

    call global%global_init ()

    libname = "integration_3"
    procname = "prc_config_a"
    
    call prepare_test_library (global, libname, 1)
    call compile_library (libname, global)

    call var_list_set_string (global%var_list, var_str ("$run_id"), &
         var_str ("integrations1"), is_known = .true.)
    call var_list_set_string (global%var_list, var_str ("$method"), &
         var_str ("unit_test"), is_known = .true.)
    call var_list_set_string (global%var_list, var_str ("$phs_method"), &
         var_str ("default"), is_known = .true.)
    call var_list_set_string (global%var_list, var_str ("$integration_method"),&
         var_str ("midpoint"), is_known = .true.)

    call var_list_set_real (global%var_list, var_str ("sqrts"),&
         1000._default, is_known = .true.)

    write (u, "(A)")  "* Create a scratch phase-space file"
    write (u, "(A)")

    u_phs = free_unit ()
    open (u_phs, file = "integrations_3.phs", &
         status = "replace", action = "write")
    call write_test_phs_file (u_phs, var_str ("prc_config_a_i1"))
    close (u_phs)

    call var_list_set_string (global%var_list, var_str ("$phs_file"),&
         var_str ("integrations_3.phs"), is_known = .true.)

    call global%it_list%init ([1], [1000])

    write (u, "(A)")  "* Integrate"
    write (u, "(A)")

    call reset_interaction_counter ()
    call integrate_process (procname, global)
    
    call global%write (u, vars = [ &
         var_str ("$phs_method"), &
         var_str ("$phs_file")])
    
    write (u, "(A)")
    write (u, "(A)")  "* Cleanup"

    call global%final ()
    call model_list_final ()
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: integrations_3"
    
  end subroutine integrations_3
  
  subroutine integrations_4 (u)
    integer, intent(in) :: u
    type(string_t) :: libname, procname
    type(rt_data_t), target :: global
    
    write (u, "(A)")  "* Test output: integrations_4"
    write (u, "(A)")  "*   Purpose: integrate test process using VAMP"
    write (u, "(A)")

    write (u, "(A)")  "* Initialize process and parameters"
    write (u, "(A)")

    call syntax_model_file_init ()

    call global%global_init ()

    libname = "integrations_4_lib"
    procname = "integrations_4"
    
    call prepare_test_library (global, libname, 1, [procname])
    call compile_library (libname, global)

    call var_list_set_string (global%var_list, var_str ("$run_id"), &
         var_str ("r1"), is_known = .true.)
    call var_list_set_string (global%var_list, var_str ("$method"), &
         var_str ("unit_test"), is_known = .true.)
    call var_list_set_string (global%var_list, var_str ("$phs_method"), &
         var_str ("single"), is_known = .true.)
    call var_list_set_string (global%var_list, var_str ("$integration_method"),&
         var_str ("vamp"), is_known = .true.)
    call var_list_set_log (global%var_list, var_str ("?use_vamp_equivalences"),&
         .false., is_known = .true.)

    call var_list_set_real (global%var_list, var_str ("sqrts"),&
         1000._default, is_known = .true.)

    call global%it_list%init ([1], [1000])

    write (u, "(A)")  "* Integrate"
    write (u, "(A)")

    call reset_interaction_counter ()
    call integrate_process (procname, global)
    
    call global%write (u, vars = [var_str ("$integration_method")])
    
    write (u, "(A)")
    write (u, "(A)")  "* Cleanup"

    call global%final ()
    call model_list_final ()
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: integrations_4"
    
  end subroutine integrations_4
  
  subroutine integrations_5 (u)
    integer, intent(in) :: u
    type(string_t) :: libname, procname
    type(rt_data_t), target :: global
    
    write (u, "(A)")  "* Test output: integrations_5"
    write (u, "(A)")  "*   Purpose: integrate test process using VAMP"
    write (u, "(A)")

    write (u, "(A)")  "* Initialize process and parameters"
    write (u, "(A)")

    call syntax_model_file_init ()

    call global%global_init ()

    libname = "integrations_5_lib"
    procname = "integrations_5"
    
    call prepare_test_library (global, libname, 1, [procname])
    call compile_library (libname, global)

    call var_list_set_string (global%var_list, var_str ("$run_id"), &
         var_str ("r1"), is_known = .true.)
    call var_list_set_string (global%var_list, var_str ("$method"), &
         var_str ("unit_test"), is_known = .true.)
    call var_list_set_string (global%var_list, var_str ("$phs_method"), &
         var_str ("single"), is_known = .true.)
    call var_list_set_string (global%var_list, var_str ("$integration_method"),&
         var_str ("vamp"), is_known = .true.)
    call var_list_set_log (global%var_list, var_str ("?use_vamp_equivalences"),&
         .false., is_known = .true.)

    call var_list_set_real (global%var_list, var_str ("sqrts"),&
         1000._default, is_known = .true.)

    call global%it_list%init ([3], [1000])

    write (u, "(A)")  "* Integrate"
    write (u, "(A)")

    call reset_interaction_counter ()
    call integrate_process (procname, global)
    
    call global%write (u, vars = [var_str ("$integration_method")])
    
    write (u, "(A)")
    write (u, "(A)")  "* Cleanup"

    call global%final ()
    call model_list_final ()
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: integrations_5"
    
  end subroutine integrations_5
  
  subroutine integrations_6 (u)
    integer, intent(in) :: u
    type(string_t) :: libname, procname
    type(rt_data_t), target :: global
    type(string_t), dimension(0) :: no_vars
    
    write (u, "(A)")  "* Test output: integrations_6"
    write (u, "(A)")  "*   Purpose: integrate test process using VAMP"
    write (u, "(A)")

    write (u, "(A)")  "* Initialize process and parameters"
    write (u, "(A)")

    call syntax_model_file_init ()

    call global%global_init ()

    libname = "integrations_6_lib"
    procname = "integrations_6"
    
    call prepare_test_library (global, libname, 1, [procname])
    call compile_library (libname, global)

    call var_list_set_string (global%var_list, var_str ("$run_id"), &
         var_str ("r1"), is_known = .true.)
    call var_list_set_string (global%var_list, var_str ("$method"), &
         var_str ("unit_test"), is_known = .true.)
    call var_list_set_string (global%var_list, var_str ("$phs_method"), &
         var_str ("single"), is_known = .true.)
    call var_list_set_string (global%var_list, var_str ("$integration_method"),&
         var_str ("vamp"), is_known = .true.)
    call var_list_set_log (global%var_list, var_str ("?use_vamp_equivalences"),&
         .false., is_known = .true.)

    call var_list_set_real (global%var_list, var_str ("sqrts"),&
         1000._default, is_known = .true.)

    call global%it_list%init ([3, 3, 3], [1000, 1000, 1000], &
         adapt = [.true., .true., .false.], &
         adapt_code = [var_str ("wg"), var_str ("g"), var_str ("")])

    write (u, "(A)")  "* Integrate"
    write (u, "(A)")

    call reset_interaction_counter ()
    call integrate_process (procname, global)
    
    call global%write (u, vars = no_vars)

    write (u, "(A)")
    write (u, "(A)")  "* Cleanup"

    call global%final ()
    call model_list_final ()
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: integrations_6"
    
  end subroutine integrations_6
  
  subroutine integrations_7 (u)
    integer, intent(in) :: u
    type(string_t) :: libname, procname
    type(rt_data_t), target :: global
    type(string_t), dimension(0) :: no_vars
    integer :: iostat, u_phs
    character(80) :: buffer
    type(string_t) :: phs_file
    logical :: exist
    
    write (u, "(A)")  "* Test output: integrations_7"
    write (u, "(A)")  "*   Purpose: integrate test process using VAMP"
    write (u, "(A)")

    write (u, "(A)")  "* Initialize process and parameters"
    write (u, "(A)")

    call syntax_model_file_init ()

    call global%global_init ()

    libname = "integrations_7_lib"
    procname = "integrations_7"
    
    call prepare_test_library (global, libname, 1, [procname])
    call compile_library (libname, global)

    call var_list_set_string (global%var_list, var_str ("$run_id"), &
         var_str ("r1"), is_known = .true.)
    call var_list_set_string (global%var_list, var_str ("$method"), &
         var_str ("unit_test"), is_known = .true.)
    call var_list_set_string (global%var_list, var_str ("$phs_method"), &
         var_str ("wood"), is_known = .true.)
    call var_list_set_string (global%var_list, var_str ("$integration_method"),&
         var_str ("vamp"), is_known = .true.)
    call var_list_set_log (global%var_list, var_str ("?use_vamp_equivalences"),&
         .true., is_known = .true.)

    call var_list_set_real (global%var_list, var_str ("sqrts"),&
         1000._default, is_known = .true.)

    call global%it_list%init ([3, 3, 3], [1000, 1000, 1000], &
         adapt = [.true., .true., .false.], &
         adapt_code = [var_str ("wg"), var_str ("g"), var_str ("")])

    write (u, "(A)")  "* Integrate"
    write (u, "(A)")

    call reset_interaction_counter ()
    call integrate_process (procname, global)
    
    call global%write (u, vars = no_vars)

    write (u, "(A)")
    write (u, "(A)")  "* Cleanup"

    call global%final ()
    call model_list_final ()
    
    write (u, "(A)")
    write (u, "(A)")  "* Generated phase-space file"
    write (u, "(A)")

    phs_file = procname // "_i1.phs"
    inquire (file = char (phs_file), exist = exist)
    if (exist) then
       u_phs = free_unit ()
       open (u_phs, file = char (phs_file), action = "read", status = "old")
       iostat = 0
       do while (iostat == 0)
          read (u_phs, "(A)", iostat = iostat)  buffer
          if (iostat == 0)  write (u, "(A)")  trim (buffer)
       end do
       close (u_phs)
    else
       write (u, "(A)")  "[file is missing]"
    end if

    write (u, "(A)")
    write (u, "(A)")  "* Test output end: integrations_7"
    
  end subroutine integrations_7
  
  subroutine integrations_8 (u)
    integer, intent(in) :: u
    type(string_t) :: libname, procname
    type(rt_data_t), target :: global
    type(flavor_t) :: flv
    
    write (u, "(A)")  "* Test output: integrations_8"
    write (u, "(A)")  "*   Purpose: integrate test process using VAMP &
         &with structure function"
    write (u, "(A)")

    write (u, "(A)")  "* Initialize process and parameters"
    write (u, "(A)")

    call syntax_model_file_init ()

    call global%global_init ()

    libname = "integrations_8_lib"
    procname = "integrations_8"
    
    call prepare_test_library (global, libname, 1, [procname])
    call compile_library (libname, global)

    call var_list_set_string (global%var_list, var_str ("$run_id"), &
         var_str ("r1"), is_known = .true.)
    call var_list_set_string (global%var_list, var_str ("$method"), &
         var_str ("unit_test"), is_known = .true.)
    call var_list_set_string (global%var_list, var_str ("$phs_method"), &
         var_str ("wood"), is_known = .true.)
    call var_list_set_string (global%var_list, var_str ("$integration_method"),&
         var_str ("vamp"), is_known = .true.)
    call var_list_set_log (global%var_list, var_str ("?use_vamp_equivalences"),&
         .true., is_known = .true.)

    call var_list_set_real (global%var_list, var_str ("sqrts"),&
         1000._default, is_known = .true.)
    call var_list_set_real (global%var_list, var_str ("ms"), &
         0._default, is_known = .true.)

    call reset_interaction_counter ()

    call flavor_init (flv, 25, global%model)
    call beam_data_init_sqrts (global%beam_data, &
         var_list_get_rval (global%var_list, var_str ("sqrts")), [flv, flv])
    call global%beam_structure%init ([1], global%var_list)
    call global%beam_structure%set_entry (1, 1, var_str ("sf_test_1"))

    write (u, "(A)")  "* Integrate"
    write (u, "(A)")

    call global%it_list%init ([1], [1000])
    call integrate_process (procname, global)
    
    call global%write (u, vars = [var_str ("ms")])

    write (u, "(A)")
    write (u, "(A)")  "* Cleanup"

    call global%final ()
    call model_list_final ()
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: integrations_8"
    
  end subroutine integrations_8
  

end module integrations
