! WHIZARD 2.2.0_alpha-1 Sep 25 2013
! 
! Copyright (C) 1999-2013 by 
!     Wolfgang Kilian <kilian@physik.uni-siegen.de>
!     Thorsten Ohl <ohl@physik.uni-wuerzburg.de>
!     Juergen Reuter <juergen.reuter@desy.de>
!     
!     with contributions by 
!     Christian Speckner <cnspeckn@googlemail.com> 
!     and  Felix Braam, Sebastian Schmidt, Daniel Wiesler 
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by 
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

module sf_escan

  use kinds, only: default !NODEP!
  use iso_varying_string, string_t => varying_string !NODEP!
  use file_utils !NODEP!
  use diagnostics !NODEP!
  use lorentz !NODEP!
  use flavors
  use helicities
  use colors
  use quantum_numbers
  use state_matrices
  use polarizations
  use interactions
  use sf_aux
  use sf_base
  
  implicit none
  private

  public :: escan_data_t

  type, extends(sf_data_t) :: escan_data_t
     private
     logical, dimension(2) :: affects_beam
     type(flavor_t), dimension(2) :: flv
     real(default), dimension(2) :: mass = 0
     real(default) :: sqrts = 0
   contains
     procedure :: init => escan_data_init
     procedure :: write => escan_data_write
     procedure :: get_n_par => escan_data_get_n_par
     procedure :: get_pdg_out => escan_data_get_pdg_out
     procedure :: allocate_sf_int => escan_data_allocate_sf_int  
  end type escan_data_t

  !!! JRR: WK please check
  type, extends (sf_int_t) :: escan_t
     type(escan_data_t), pointer :: data => null ()
     real(default), dimension(2) :: x = 0
     real(default), dimension(2) :: q = 0
   contains
     procedure :: type_string => escan_type_string
     procedure :: write => escan_write
     procedure :: complete_kinematics => escan_complete_kinematics
     procedure :: inverse_kinematics => escan_inverse_kinematics
     procedure :: init => escan_init
     procedure :: apply => escan_apply
  end type escan_t 
  

contains

  subroutine escan_data_init (data, affects_beam, flv_in, sqrts)
    class(escan_data_t), intent(out) :: data
    logical, dimension(2), intent(in) :: affects_beam
    type(flavor_t), dimension(2), intent(in) :: flv_in
    real(default), intent(in) :: sqrts
    data%affects_beam = affects_beam
    data%flv = flv_in
    data%mass = flavor_get_mass (data%flv)
    data%sqrts = sqrts
  end subroutine escan_data_init

  subroutine escan_data_write (data, unit, verbose) !, md5)
    class(escan_data_t), intent(in) :: data
    integer, intent(in), optional :: unit
    logical, intent(in), optional :: verbose
    integer :: u
    !!! JRR: WK please check: do we need this here?
!    logical, intent(in), optional :: md5
    u = output_unit (unit);  if (u < 0)  return
    write (u, "(1x,A)") "Energy-scan data:"
    write (u, "(3x,A,A,A,A)") &
         "  prt_in  = ", char (flavor_get_name (data%flv(1))), &
         ", ", char (flavor_get_name (data%flv(2)))    
  end subroutine escan_data_write

  function escan_data_get_n_par (data) result (n)
    !!! JRR: WK please check
    !!! What about correlation of x values?
    class(escan_data_t), intent(in) :: data
    integer :: n
    n = count (data%affects_beam)
  end function escan_data_get_n_par
  
!!! JRR: WK please check
  function escan_data_get_pdg_out (data) result (pdg_out)
    class(escan_data_t), intent(in) :: data
    integer, dimension(:), allocatable :: pdg_out
    integer :: i, n
    n = 2
    allocate (pdg_out (n))
    do i = 1, n
       pdg_out(i) = flavor_get_pdg (data%flv(i))
    end do
  end function escan_data_get_pdg_out
  
  subroutine escan_data_allocate_sf_int (data, sf_int)
    class(escan_data_t), intent(in) :: data
    class(sf_int_t), intent(inout), allocatable :: sf_int
    allocate (escan_t :: sf_int)
  end subroutine escan_data_allocate_sf_int
  
  function escan_type_string (object) result (string)
    !!! JRR: WK please check
    class(escan_t), intent(in) :: object
    type(string_t) :: string
    if (associated (object%data)) then
       string = "Escan: energy scan" 
    else
       string = "Escan: [undefined]"
    end if
  end function escan_type_string
  
  subroutine escan_write (object, unit)
    !!! JRR: WK please check
    !!! Guess these variables do not exist for CIRCE1 (?)
    class(escan_t), intent(in) :: object
    integer, intent(in), optional :: unit
    integer :: u
    u = output_unit (unit)
    if (associated (object%data)) then
       call object%data%write (u)
       if (object%status >= SF_DONE_KINEMATICS) then
          write (u, "(1x,A)")  "SF parameters:"
          write (u, "(3x,A,ES17.10)")  "x =", object%x
          if (object%status >= SF_FAILED_EVALUATION) then
             write (u, "(3x,A,ES17.10)")  "Q =", object%q
          end if
       end if
       call object%base_write (u)
    else
       write (u, "(1x,A)")  "Energy scan data: [undefined]"
    end if
  end subroutine escan_write
    
  subroutine escan_complete_kinematics (sf_int, x, f, r, map)
    class(escan_t), intent(inout) :: sf_int
    real(default), dimension(:), intent(out) :: x
    real(default), intent(out) :: f
    real(default), dimension(:), intent(in) :: r
    logical, intent(in) :: map
    real(default) :: xb1
    if (map) then
       call msg_fatal ("Energy scan: map flag not supported")
    else
       x(1) = r(1)
       f = 1
    end if
    xb1 = 1 - x(1)
    call sf_int%split_momentum (x, xb1)
    select case (sf_int%status)
    case (SF_DONE_KINEMATICS)
       sf_int%x = x(1)
    case (SF_FAILED_KINEMATICS)
       sf_int%x = 0
       f = 0
    end select
  end subroutine escan_complete_kinematics

  subroutine escan_inverse_kinematics (sf_int, x, f, r, map, set_momenta)
    class(escan_t), intent(inout) :: sf_int
    real(default), dimension(:), intent(in) :: x
    real(default), intent(out) :: f
    real(default), dimension(:), intent(out) :: r
    logical, intent(in) :: map
    logical, intent(in), optional :: set_momenta
    real(default) :: xb1
    logical :: set_mom
    set_mom = .false.;  if (present (set_momenta))  set_mom = set_momenta
    if (map) then
       call msg_fatal ("Energy scan: map flag not supported")
    else
       r(1) = x(1)
       f = 1
    end if
    xb1 = 1 - x(1)
    if (set_mom) then
       call sf_int%split_momentum (x, xb1)
       select case (sf_int%status)
       case (SF_DONE_KINEMATICS)
          sf_int%x = x(1)
       case (SF_FAILED_KINEMATICS)
          sf_int%x = 0
          f = 0
       end select
    end if
  end subroutine escan_inverse_kinematics

  subroutine escan_init (sf_int, data)
    class(escan_t), intent(out) :: sf_int
    class(sf_data_t), intent(in), target :: data
    type(quantum_numbers_mask_t), dimension(4) :: mask
    integer, dimension(4) :: hel_lock
    real(default), dimension(0) :: null_array    
    type(quantum_numbers_t), dimension(4) :: qn_fc, qn_hel, qn
    type(polarization_t) :: pol1, pol2
    type(state_iterator_t) :: it_hel1, it_hel2
    integer :: i
    select type (data)
    type is (escan_data_t)
       if (all (data%affects_beam)) then
          hel_lock = (/ 3, 4, 1, 2 /)
          call sf_int%base_init (mask, data%mass**2, &
               null_array, data%mass**2, hel_lock = hel_lock)
       else if (data%affects_beam(1)) then
          hel_lock = (/ 2, 1, 0, 0 /)
          call sf_int%base_init (mask(1:2), [data%mass(1)**2], &
               null_array, [data%mass(1)**2], hel_lock = hel_lock(1:2))
       else if (data%affects_beam(2)) then
          hel_lock = (/ 2, 1, 0, 0 /)
          !!! JRR: WK please check
          !!! Seems to be a cut-and-paste error, is it really
          !!! identical to the case affects_beam(1) ??
          call sf_int%base_init (mask(1:2), [data%mass(2)**2], &
               null_array, [data%mass(2)**2], hel_lock = hel_lock(1:2))
       end if
       sf_int%data => data       
       do i = 1, 2
          call quantum_numbers_init (qn_fc(i), &
               flv = data%flv(i), col = color_from_flavor (data%flv(i)))
          call quantum_numbers_init (qn_fc(i+2), &
               flv = data%flv(i), col = color_from_flavor (data%flv(i)))
       end do
       call polarization_init_generic (pol1, data%flv(1))
       call state_iterator_init (it_hel1, pol1%state)
       do while (state_iterator_is_valid (it_hel1))
          qn_hel(1:1) = state_iterator_get_quantum_numbers (it_hel1)
          qn_hel(3:3) = state_iterator_get_quantum_numbers (it_hel1)
          call polarization_init_generic (pol2, data%flv(2))
          call state_iterator_init (it_hel2, pol2%state)
          do while (state_iterator_is_valid (it_hel2))
             qn_hel(2:2) = state_iterator_get_quantum_numbers (it_hel2)
             qn_hel(4:4) = state_iterator_get_quantum_numbers (it_hel2)
             qn = qn_hel .merge. qn_fc
             if (all (data%affects_beam)) then
                call interaction_add_state (sf_int%interaction_t, qn)
             else if (data%affects_beam(1)) then
                call interaction_add_state (sf_int%interaction_t, qn((/1,3/)))
             else if (data%affects_beam(2)) then
                call interaction_add_state (sf_int%interaction_t, qn((/2,4/)))
             end if
             call state_iterator_advance (it_hel2)
          end do
          call polarization_final (pol2)
          call state_iterator_advance (it_hel1)
       end do
       call polarization_final (pol2)
       call interaction_freeze (sf_int%interaction_t)
       !!! JRR: WK please check
       call sf_int%set_incoming ([1,2])
       call sf_int%set_outgoing ([3,4])
       sf_int%status = SF_INITIAL
    end select
  end subroutine escan_init
    
  subroutine escan_apply (sf_int, scale) !!! r, data)
    !!! JRR: WK please check, different from the PDFs
    class(escan_t), intent(inout) :: sf_int
    !!! Dummy
    real(default), intent(in) :: scale
    !!! real(default), dimension(1), intent(in) :: r
    !!! type(escan_data_t), intent(in) :: data
    !!! type(vector4_t), dimension(2) :: k
    real(default) :: f, xx
    !!! real(default), dimension(2) :: x
    !!! type(splitting_data_t), dimension(2) :: sd
    !!! type(vector4_t), dimension(4) :: q
    associate (data => sf_int%data)
    !!! k(1) = interaction_get_momentum (int, 1)
    !!! if (all (data%affects_beam)) then
    !!!    k(2) = interaction_get_momentum (int, 2)
    !!! else
    !!!    k(2) = k(1)
    !!! end if
    !!! call strfun (f, xx, r(1), data)
    !!! if (all (data%affects_beam)) then
    !!!    x = sqrt (xx)
    !!! else
    !!!    where (data%affects_beam)
    !!!       x = xx
    !!!    elsewhere
    !!!       x = 1
    !!!    end where
    !!! end if
    !!! call interaction_set_matrix_element (int, cmplx (f, kind=default))
    !!! sd = new_splitting_data (k, data%mass**2, 0._default, data%mass)
    !!! call splitting_set_t_bounds (sd, x, 1-x)
    !!! call splitting_set_collinear (sd)
    !!! q((/1, 3/)) = split_momentum (k(1), sd(1))
    !!! q((/2, 4/)) = split_momentum (k(2), sd(2))
    !!! if (all (data%affects_beam)) then
    !!!    call interaction_set_momenta (int, q(3:4), outgoing=.true.)
    !!! else if (data%affects_beam(1)) then
    !!!    call interaction_set_momenta (int, q(3:3), outgoing=.true.)
    !!! else if (data%affects_beam(2)) then
    !!!    call interaction_set_momenta (int, q(4:4), outgoing=.true.)
    !!! end if
    end associate
    sf_int%status = SF_EVALUATED
  end subroutine escan_apply


end module sf_escan
