! WHIZARD 2.2.0_alpha-1 Sep 25 2013
! 
! Copyright (C) 1999-2013 by 
!     Wolfgang Kilian <kilian@physik.uni-siegen.de>
!     Thorsten Ohl <ohl@physik.uni-wuerzburg.de>
!     Juergen Reuter <juergen.reuter@desy.de>
!     
!     with contributions by 
!     Christian Speckner <cnspeckn@googlemail.com> 
!     and  Felix Braam, Sebastian Schmidt, Daniel Wiesler 
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by 
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

module sf_ewa

  use kinds, only: default !NODEP!
  use iso_varying_string, string_t => varying_string !NODEP!
  use file_utils !NODEP!
  use constants, only: pi !NODEP!
  use diagnostics !NODEP!
  use lorentz !NODEP!
  use unit_tests
  use os_interface
  use models
  use flavors
  use colors
  use quantum_numbers
  use state_matrices
  use polarizations
  use interactions
  use sf_aux
  use sf_base

  implicit none
  private

  public :: ewa_data_t
  public :: sf_ewa_test

  integer, parameter :: NONE = 0
  integer, parameter :: ZERO_QMIN = 1
  integer, parameter :: Q_MAX_TOO_SMALL = 2
  integer, parameter :: ZERO_XMIN = 3

  type, extends(sf_data_t) :: ewa_data_t
     private
     type(model_t), pointer :: model => null ()
     type(flavor_t) :: flv
     real(default) :: pt_max
     real(default) :: sqrts
     real(default) :: x_min
     real(default) :: x_max
     real(default) :: mass
     real(default) :: q_min
     real(default) :: cv
     real(default) :: ca
     real(default) :: costhw
     real(default) :: sinthw
     real(default) :: mW
     real(default) :: mZ
     real(default) :: coeff
     logical :: keep_momentum
     logical :: keep_energy     
     integer :: id = 0 
     integer :: error = 0
     logical :: map = .true.
   contains
     procedure :: init => ewa_data_init
     procedure :: check => ewa_data_check
     !!! JRR: WK please check. This has to be set somewhere in the
     !!! new structure. I have no clue, where. :(  
     procedure :: set_id => ewa_set_id
     procedure :: write => ewa_data_write
     procedure :: get_n_par => ewa_data_get_n_par
     procedure :: get_pdg_out => ewa_data_get_pdg_out
     procedure :: allocate_sf_int => ewa_data_allocate_sf_int  
  end type ewa_data_t

  !!! JRR: WK please check
  type, extends (sf_int_t) :: ewa_t
     type(ewa_data_t), pointer :: data => null ()
     real(default) :: x = 0
     real(default) :: q = 0
   contains
     procedure :: type_string => ewa_type_string
     procedure :: write => ewa_write
     procedure :: init => ewa_init
     procedure :: complete_kinematics => ewa_complete_kinematics
     procedure :: inverse_kinematics => ewa_inverse_kinematics
     procedure :: apply => ewa_apply
  end type ewa_t 
  

contains

  subroutine ewa_data_init (data, model, flv, x_min, q_min, pt_max, &
        sqrts, keep_momentum, keep_energy, mass)
    class(ewa_data_t), intent(inout) :: data
    type(model_t), intent(in), target :: model
    type(flavor_t), intent(in) :: flv
    real(default), intent(in) :: x_min, q_min, pt_max, sqrts
    logical, intent(in) :: keep_momentum, keep_energy
    real(default), intent(in), optional :: mass
    real(default) :: g, ee, sinthw
    data%model => model
    data%flv = flv
    data%pt_max = pt_max
    data%sqrts = sqrts
    data%x_min = x_min
    data%x_max = 1
    if (data%x_min == 0) then
       data%error = ZERO_XMIN;  return
    end if    
    select case (char (model_get_name (data%model)))
    case ("QCD","QED","Test")
       call msg_fatal ("EWA structure function not available for model " &
            // char (model_get_name (data%model)) // ".")
    end select     
    ee = model_get_parameter_value (data%model, var_str ("ee"))
    data%sinthw = model_get_parameter_value (data%model, var_str ("sw"))    
    data%costhw = model_get_parameter_value (data%model, var_str ("cw"))        
    data%mZ = model_get_parameter_value (data%model, var_str ("mZ"))
    data%mW = model_get_parameter_value (data%model, var_str ("mW"))    
    if (data%sinthw /= 0) then
       g = ee / data%sinthw
    else
       call msg_fatal ("Vanishing value of sin(theta_w).")
    end if
    data%cv = g / 2._default
    data%ca = g / 2._default   
    data%coeff = 1._default / (8._default * PI**2)
    data%keep_momentum = keep_momentum
    data%keep_energy = keep_energy
    if (present (mass)) then
       data%mass = mass
    else
       data%mass = flavor_get_mass (flv)
    end if
  end subroutine ewa_data_init

  subroutine ewa_data_check (data)
    class(ewa_data_t), intent(in) :: data
    select case (data%error)
    case (ZERO_QMIN)
       call msg_fatal (" EWA: Particle mass is zero")
    case (Q_MAX_TOO_SMALL)
       call msg_fatal (" EWA: Particle mass exceeds Qmax")
    case (ZERO_XMIN)
       call msg_fatal (" EWA: x_min must be larger than zero")
    end select
  end subroutine ewa_data_check

  subroutine ewa_set_id (data, id)
    class(ewa_data_t), intent(inout) :: data
    integer, intent(in) :: id
    data%id = id
  end subroutine ewa_set_id 

  subroutine ewa_data_write (data, unit, verbose) !, md5)
    class(ewa_data_t), intent(in) :: data
    integer, intent(in), optional :: unit
    logical, intent(in), optional :: verbose
    integer :: u
    !!! JRR: WK please check: still needed?
!    logical, intent(in), optional :: md5
    u = output_unit (unit);  if (u < 0)  return
    write (u, "(1x,A)") "EWA data:"
    if (flavor_is_defined (data%flv)) then
       write (u, "(3x,A)", advance="no") "  flavor   =  "
       call flavor_write (data%flv, u); write (u, *)    
       write (u, "(3x,A,ES19.12)") "  x_min     = ", data%x_min
       write (u, "(3x,A,ES19.12)") "  x_max     = ", data%x_max
       write (u, "(3x,A,ES19.12)") "  pt_max    = ", data%pt_max
       write (u, "(3x,A,ES19.12)") "  sqrts     = ", data%sqrts    
       write (u, "(3x,A,ES19.12)") "  mass      = ", data%mass
       write (u, "(3x,A,ES19.12)") "  cv        = ", data%cv
       write (u, "(3x,A,ES19.12)") "  ca        = ", data%ca
       write (u, "(3x,A,ES19.12)") "  coeff     = ", data%coeff
       write (u, "(3x,A,ES19.12)") "  costhw    = ", data%costhw
       write (u, "(3x,A,ES19.12)") "  sinthw    = ", data%sinthw    
       write (u, "(3x,A,ES19.12)") "  mZ        = ", data%mZ
       write (u, "(3x,A,ES19.12)") "  mW        = ", data%mW
       write (u, "(3x,A,L2)")      "  keep_mom. = ", data%keep_momentum
       write (u, "(3x,A,L2)")      "  keep_en.  = ", data%keep_energy 
       write (u, "(3x,A,L2)")      "  map       = ", data%map
    else
       write (u, "(3x,A)") "[undefined]"
    end if       
  end subroutine ewa_data_write

  function ewa_data_get_n_par (data) result (n)
    class(ewa_data_t), intent(in) :: data
    integer :: n
    if (data%keep_energy .or. data%keep_momentum) then
       n = 3
    else
       n = 1
    end if
  end function ewa_data_get_n_par
  
!!! JRR: WK please check
  function ewa_data_get_pdg_out (data) result (pdg_out)
    class(ewa_data_t), intent(in) :: data
    integer, dimension(:), allocatable :: pdg_out
    real(default) :: t3
    logical :: up_type
    integer :: i, n
    n = 1
    allocate (pdg_out (n))
    t3 = - flavor_get_isospin (data%flv)
    up_type = (t3 > 0)
    select case (data%id)
       case(23)
          pdg_out(1) = flavor_get_pdg (data%flv)
       case(24)
          if (up_type) then
             pdg_out(1) = flavor_get_pdg (data%flv) - 1
          else
             pdg_out(1) = flavor_get_pdg (data%flv) + 1
          end if
    end select
  end function ewa_data_get_pdg_out
  
  subroutine ewa_data_allocate_sf_int (data, sf_int)
    class(ewa_data_t), intent(in) :: data
    class(sf_int_t), intent(inout), allocatable :: sf_int
    allocate (ewa_t :: sf_int)
  end subroutine ewa_data_allocate_sf_int
  
  function ewa_type_string (object) result (string)
    class(ewa_t), intent(in) :: object
    type(string_t) :: string
    if (associated (object%data)) then
       string = "EWA: equivalent W/Z approx." 
    else
       string = "EWA: [undefined]"
    end if
  end function ewa_type_string
  
  subroutine ewa_write (object, unit)
    !!! JRR: WK please check
    class(ewa_t), intent(in) :: object
    integer, intent(in), optional :: unit
    integer :: u
    u = output_unit (unit)
    if (associated (object%data)) then
       call object%data%write (u)
       if (object%status >= SF_DONE_KINEMATICS) then
          write (u, "(1x,A)")  "SF parameters:"
          write (u, "(3x,A,ES17.10)")  "x =", object%x
          if (object%status >= SF_FAILED_EVALUATION) then
             write (u, "(3x,A,ES17.10)")  "Q =", object%q
          end if
       end if
       call object%base_write (u)
    else
       write (u, "(1x,A)")  "EWA data: [undefined]"
    end if
  end subroutine ewa_write
    
  subroutine ewa_init (sf_int, data)
    !!! JRR: WK please check
    class(ewa_t), intent(out) :: sf_int
    class(sf_data_t), intent(in), target :: data
    type(quantum_numbers_mask_t), dimension(3) :: mask
    integer, dimension(3) :: hel_lock
    type(polarization_t) :: pol
    type(quantum_numbers_t), dimension(1) :: qn_fc, qn_hel, qn_fc_fin
    type(flavor_t) :: flv_z, flv_wp, flv_wm, flv_down, flv_up
    type(quantum_numbers_t) :: qn_z, qn_wp, qn_wm, qn, qn_down, &
         qn_up
    type(state_iterator_t) :: it_hel
    real(default) :: q, t3  
    real(default) :: m_out
    integer :: i, isospin, pdg
    logical :: up_type
    select type (data)
    type is (ewa_data_t)   
       isospin = flavor_get_isospin_type (data%flv)
       pdg = flavor_get_pdg (data%flv) 
       if (abs(isospin) /= 2 .or. abs(pdg) > 16 .or. pdg == 0 .or. pdg == 9) then
          call msg_fatal ("EWA structure function only accessible for " &
               // "SM quarks and leptons.")
       end if
       q = - flavor_get_charge (data%flv)        
       t3 = - flavor_get_isospin (data%flv)
       up_type = (t3 > 0)
       mask = new_quantum_numbers_mask (.false., .false., &
            mask_h = (/ .false., .false., .true. /))
       hel_lock = (/ 2, 1, 0 /)
       select case (data%id)
       case (23)
          !!! Z boson, flavor is not changing    
          call sf_int%base_init (mask, [data%mass**2], [data%mass**2], &
               [data%mZ**2], hel_lock = hel_lock)
          sf_int%data => data          
          call flavor_init (flv_z, Z_BOSON, data%model)
          call quantum_numbers_init (qn_z, flv_z)
          call polarization_init_generic (pol, data%flv)
          call quantum_numbers_init (qn_fc(1), &
               flv = data%flv, col = color_from_flavor (data%flv, 1))
          call state_iterator_init (it_hel, pol%state)
          do while (state_iterator_is_valid (it_hel))
             qn_hel = state_iterator_get_quantum_numbers (it_hel)
             qn = qn_hel(1) .merge. qn_fc(1)
             call interaction_add_state &
                  (sf_int%interaction_t, (/ qn, qn, qn_z /))
             call state_iterator_advance (it_hel)
          end do
          call polarization_final (pol)
       case (24)    
          if (up_type) then            
             !!! W+, flavor changing
             call interaction_init (sf_int%interaction_t, &
                  1, 0, 2, mask=mask, hel_lock=hel_lock, set_relations=.true.)             
             call flavor_init (flv_down, pdg - 1, data%model)
             call flavor_init (flv_wp, W_BOSON, data%model)
             m_out = flavor_get_mass (flv_down)
             call sf_int%base_init (mask, [data%mass**2], [m_out**2], &
                  [data%mW**2], hel_lock = hel_lock)
             sf_int%data => data                       
             call quantum_numbers_init (qn_wp, flv_wp)
             call polarization_init_generic (pol, data%flv)
             call quantum_numbers_init (qn_fc(1), &
                  flv = data%flv, col = color_from_flavor (data%flv, 1))
             call quantum_numbers_init (qn_fc_fin(1), &
                  flv = flv_down, col = color_from_flavor (flv_down))
             call state_iterator_init (it_hel, pol%state)
             do while (state_iterator_is_valid (it_hel))
                qn_hel = state_iterator_get_quantum_numbers (it_hel)
                qn = qn_hel(1) .merge. qn_fc(1)
                qn_down = qn_hel(1) .merge. qn_fc_fin(1)           
                call interaction_add_state &
                     (sf_int%interaction_t, (/ qn, qn_down, qn_wp /))
                call state_iterator_advance (it_hel)
             end do
             call polarization_final (pol)    
          else
             !!! W-, flavor changing
             call flavor_init (flv_up, pdg + 1, data%model)
             call flavor_init (flv_wm, - W_BOSON, data%model)
             m_out = flavor_get_mass (flv_up)
             call sf_int%base_init (mask, [data%mass**2], [m_out**2], &
                  [data%mW**2], hel_lock = hel_lock)
             sf_int%data => data                                    
             call quantum_numbers_init (qn_wm, flv_wm)
             call polarization_init_generic (pol, data%flv)
             call quantum_numbers_init (qn_fc(1), &
                  flv = data%flv, col = color_from_flavor (data%flv, 1))
             call quantum_numbers_init (qn_fc_fin(1), &
                  flv = flv_up, col = color_from_flavor (flv_up))
             call state_iterator_init (it_hel, pol%state)
             do while (state_iterator_is_valid (it_hel))
                qn_hel = state_iterator_get_quantum_numbers (it_hel)
                qn = qn_hel(1) .merge. qn_fc(1)
                qn_up = qn_hel(1) .merge. qn_fc_fin(1)           
                call interaction_add_state &
                     (sf_int%interaction_t, (/ qn, qn_up, qn_wm /))
                call state_iterator_advance (it_hel)
             end do
             call polarization_final (pol)    
          end if
       case default
          call msg_fatal ("EWA initialization failed: wrong particle type.")
       end select
       call interaction_freeze (sf_int%interaction_t)
       !!! JRR: WK please check
       !!! is that the right place to set the keep_ flags
       if (data%keep_momentum) then
          if (data%keep_energy) then
             call msg_fatal ("EWA: momentum and energy" // &
                  "cannot be simultaneously conserved.")
          else          
             sf_int%on_shell_mode = KEEP_MOMENTUM
          end if
       else 
          if (data%keep_energy) then
             sf_int%on_shell_mode = KEEP_ENERGY
          end if
       end if
       call sf_int%set_incoming ([1])
       call sf_int%set_radiated ([2])
       call sf_int%set_outgoing ([3])
       sf_int%status = SF_INITIAL
    end select
  end subroutine ewa_init
    
  subroutine ewa_complete_kinematics (sf_int, x, f, r, map)
    !!! JRR: WK please check
    class(ewa_t), intent(inout) :: sf_int
    real(default), dimension(:), intent(out) :: x
    real(default), intent(out) :: f
    real(default), dimension(:), intent(in) :: r
    logical, intent(in) :: map
    real(default) :: xb1, x0, x1, lx0, lx1, lx
    if (sf_int%data%keep_momentum .or. sf_int%data%keep_energy) then
       select case (sf_int%data%id)
       case (23)
          x0 = max (sf_int%data%x_min, sf_int%data%mz/sf_int%q)
       case (24)
          x0 = max (sf_int%data%x_min, sf_int%data%mw/sf_int%q)
       end select
    else 
       x0 = sf_int%data%x_min
    end if
    x1 = sf_int%data%x_max
    if ( x0 >= x1) then
       f = 0
       sf_int%status = SF_FAILED_KINEMATICS
       return
    end if
    if (map) then
       lx0 = log (x0)
       lx1 = log (x1)
       lx = lx1 * r(1) + lx0 * (1 - r(1))
       x(1) = exp(lx)       
       !!! JRR: WK please check, factor x would be part of Jacobian
       !!! but cancels 1/x in structure function. I do the cancellation
       !!! explicitly, depending on data%map
       f = lx1 - lx0    !!! x(1) * (lx1 - lx0)
    else
       x(1) = r(1)
       f = 1
    end if
    xb1 = 1 - x(1)
    if (size(x) == 3)  x(2:3) = r(2:3)
    !!! JRR: WK please check
    !!! in kinematics we have to set the map flag in order
    !!! to be available in apply
    sf_int%data%map = map    
    call sf_int%split_momentum (x, xb1)
    select case (sf_int%status)
    case (SF_DONE_KINEMATICS)
       sf_int%x = x(1)
    case (SF_FAILED_KINEMATICS)
       sf_int%x = 0
       f = 0
    end select
  end subroutine ewa_complete_kinematics

  subroutine ewa_inverse_kinematics (sf_int, x, f, r, map, set_momenta)
    !!! JRR: WK please check
    class(ewa_t), intent(inout) :: sf_int
    real(default), dimension(:), intent(in) :: x
    real(default), intent(out) :: f
    real(default), dimension(:), intent(out) :: r
    logical, intent(in) :: map
    logical, intent(in), optional :: set_momenta
    real(default) :: xb1, x0, x1, lx0, lx1, lx
    logical :: set_mom
    set_mom = .false.;  if (present (set_momenta))  set_mom = set_momenta
    if (sf_int%data%keep_momentum .or. sf_int%data%keep_energy) then
       select case (sf_int%data%id)
       case (23)
          x0 = max (sf_int%data%x_min, sf_int%data%mz/sf_int%q)
       case (24)
          x0 = max (sf_int%data%x_min, sf_int%data%mw/sf_int%q)
       end select
    else 
       x0 = sf_int%data%x_min
    end if
    x1 = sf_int%data%x_max
    if (map) then
       lx0 = log (x0)
       lx1 = log (x1)
       r(1) = (log(x(1)) - lx0) / (lx1 - lx0)
       !!! JRR: WK please check, factor x would be part of Jacobian
       !!! but cancels 1/x in structure function. I do the cancellation
       !!! explicitly, depending on data%map       
       f = lx1 - lx0   !!! x(1) * (lx1 - lx0)
    else
       r(1) = x(1)
       f = 1
    end if
    xb1 = 1 - x(1)
    if (size(r) == 3)  r(2:3) = x(2:3)
    if (set_mom) then
       call sf_int%split_momentum (x, xb1)
       select case (sf_int%status)
       case (SF_DONE_KINEMATICS)
          sf_int%x = x(1)
       case (SF_FAILED_KINEMATICS)
          sf_int%x = 0
          f = 0
       end select
    end if
  end subroutine ewa_inverse_kinematics

  subroutine ewa_apply (sf_int, scale) !! r, ewa_data, no_map)
    class(ewa_t), intent(inout) :: sf_int
    real(default), intent(in) :: scale
    real(default) :: x, xb, pt2, c1, c2, q, t3
    real(default) :: cv, ca
    real(default) :: f, fm, fp, fL
    associate (data => sf_int%data)     
      x = sf_int%x      
      xb = 1 - x
      fm = 0
      fp = 0
      fL = 0
      !!! JRR: WK please check: should be correct. The structure
      !!! function depends on mapped variables in the mapped case
      !!! and the true r values (constrained to interval [xmin, xmax])
      !!! for the unmapped case. Jacobian is taken into account in
      !!! complete_kinematics. Is this true also for the qminsq bound
      !!! or do we have to use the unmapped x values there?      
      !!! Furthermore: in W 2.1.1 fm, fp, fL have not been used...
      pt2 = min ((data%pt_max)**2, (xb * data%sqrts / 2)**2)
      select case (data%id)
      case (23)
         !!! Z boson structure function
         c1 = log (1 + pt2 / (xb * (data%mZ)**2))
         c2 = 1 / (1 + (xb * (data%mZ)**2) / pt2)
         if (flavor_is_antiparticle (data%flv)) then
            q = flavor_get_charge (data%flv)
            t3 = flavor_get_isospin (data%flv)
         else     
            q = - flavor_get_charge (data%flv)        
            t3 = - flavor_get_isospin (data%flv)        
         end if
         cv = data%cv * (t3 - 2._default * q * data%sinthw**2) / data%costhw
         ca = data%ca *  t3 / data%costhw    
         if (flavor_is_antiparticle (data%flv)) ca = - ca
         fm = data%coeff * &
              ((cv + ca)**2 + ((cv - ca) * xb)**2) / 2 * (c1 - c2)
         fp = data%coeff * &
              ((cv - ca)**2 + ((cv + ca) * xb)**2) / 2 * (c1 - c2)    
         fL = data%coeff * &
              (cv**2 + ca**2) * 2 * xb * c2
         if (.not. data%map) then
            fm = fm / x
            fp = fm / x
            fL = fL / x
         end if
         f = fp + fm + fL
         if (f /= 0) then
            fp = fp / f
            fm = fm / f
            fL = fL / f
         end if
      case (24)
         !!! W boson structure function
         c1 = log (1 + pt2 / (xb * (data%mW)**2))
         c2 = 1 / (1 + (xb * (data%mW)**2) / pt2)
         cv = data%cv / sqrt(2._default)
         ca = data%ca / sqrt(2._default)
         if (flavor_is_antiparticle (data%flv)) ca = - ca        
         fm = data%coeff * &
              ((cv + ca)**2 + ((cv - ca) * xb)**2) / 2 * (c1 - c2)
         fp = data%coeff * &
              ((cv - ca)**2 + ((cv + ca) * xb)**2) / 2 * (c1 - c2)    
         fL = data%coeff * &
              (cv**2 + ca**2) * 2 * xb * c2    
         if (.not. data%map) then
            fm = fm / x
            fp = fm / x
            fL = fL / x
         end if
         f = fp + fm + fL
         if (f /= 0) then
            fp = fp / f
            fm = fm / f
            fL = fL / f
         end if
      end select
    end associate
    call interaction_set_matrix_element &
         (sf_int%interaction_t, cmplx (f, kind=default))
    sf_int%status = SF_EVALUATED
  end subroutine ewa_apply


  subroutine sf_ewa_test (u, results)
    integer, intent(in) :: u
    type(test_results_t), intent(inout) :: results
    !!! JRR: WK please check
    !!! there are probably more tests necessary. e.g. to see whether
    !!! integrals with and without mapping are the same, right?
    call test (sf_ewa_1, "sf_ewa_1", &
         "structure function configuration", &
         u, results)
    call test (sf_ewa_2, "sf_ewa_2", &
         "structure function instance", &
         u, results)
  end subroutine sf_ewa_test
  
  subroutine sf_ewa_1 (u)
    integer, intent(in) :: u
    type(os_data_t) :: os_data
    type(model_t), pointer :: model
    type(flavor_t) :: flv
    class(sf_data_t), allocatable :: data
    
    write (u, "(A)")  "* Test output: sf_ewa_1"
    write (u, "(A)")  "*   Purpose: initialize and display &
         &test structure function data"
    write (u, "(A)")
    
    write (u, "(A)")  "* Create empty data object"
    write (u, "(A)")

    call os_data_init (os_data)
    call syntax_model_file_init ()
    call model_list_read_model (var_str ("SM"), &
         var_str ("SM.mdl"), os_data, model)
    call flavor_init (flv, 2, model)

    allocate (ewa_data_t :: data)
    call data%write (u)

    write (u, "(A)")
    write (u, "(A)")  "* Initialize for Z boson"
    write (u, "(A)")

    select type (data)
    type is (ewa_data_t)
       call data%init (model, flv, 0.01_default, 10._default, &
            500._default, 5000._default, .false., .false., 5._default)
       call data%set_id (23)
    end select

    call data%write (u)

    write (u, "(A)")

    write (u, "(1x,A)")  "Outgoing particle codes:"
    write (u, "(2x,99(1x,I0))")  data%get_pdg_out ()
        
    write (u, "(A)")
    write (u, "(A)")  "* Initialize for W boson"
    write (u, "(A)")
    
    select type (data)
    type is (ewa_data_t)
       call data%init (model, flv, 0.01_default, 10._default, &
            500._default, 5000._default, .false., .false., 5._default)
       call data%set_id (24)
    end select

    call data%write (u)

    write (u, "(A)")

    write (u, "(1x,A)")  "Outgoing particle codes:"
    write (u, "(2x,99(1x,I0))")  data%get_pdg_out ()
            
    call model_list_final ()
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: sf_ewa_1"

  end subroutine sf_ewa_1

  subroutine sf_ewa_2 (u)
    integer, intent(in) :: u
    type(os_data_t) :: os_data
    type(model_t), pointer :: model
    type(flavor_t) :: flv
    class(sf_data_t), allocatable, target :: data
    class(sf_int_t), allocatable :: sf_int
    type(vector4_t) :: k
    type(vector4_t), dimension(2) :: q
    real(default) :: E
    real(default), dimension(:), allocatable :: r, x
    real(default) :: f, s
    
    write (u, "(A)")  "* Test output: sf_ewa_2"
    write (u, "(A)")  "*   Purpose: initialize and fill &
         &test structure function object"
    write (u, "(A)")
    
    write (u, "(A)")  "* Initialize configuration data"
    write (u, "(A)")

    call os_data_init (os_data)
    call syntax_model_file_init ()
    call model_list_read_model (var_str ("SM"), &
         var_str ("SM.mdl"), os_data, model)
    call flavor_init (flv, 2, model)

    call reset_interaction_counter ()
    
    allocate (ewa_data_t :: data)
    select type (data)
    type is (ewa_data_t)
       call data%init (model, flv, 0.01_default, 10._default, &
            500._default, 5000._default, .false., .false., 5._default)       
       call data%set_id (24)
    end select
       
    write (u, "(A)")  "* Initialize structure-function object"
    write (u, "(A)")
    
    call data%allocate_sf_int (sf_int)
    call sf_int%init (data)
    call sf_int%set_beam_index ([1])
    
    call sf_int%write (u)

    write (u, "(A)")
    write (u, "(A)")  "* Initialize incoming momentum with E=500"
    write (u, "(A)")
    E = 500
    k = vector4_moving (E, sqrt (E**2 - flavor_get_mass (flv)**2), 3)
    call pacify (k, 1e-10_default)
    call vector4_write (k, u)
    call sf_int%seed_kinematics ([k])
    
    write (u, "(A)")
    write (u, "(A)")  "* Set kinematics for x=0.5, no EWA mapping, collinear"
    write (u, "(A)")
    
    allocate (r (data%get_n_par ()))
    allocate (x (size (r)))
    
    r = 0.5_default
    select type (sf_int)
    type is (ewa_t)
       sf_int%q = 1000._default
    end select
    call sf_int%complete_kinematics (x, f, r, map=.false.)
    call sf_int%write (u)
    
    write (u, "(A)")
    write (u, "(A,9(1x,F12.9))")  "x =", x
    write (u, "(A,9(1x,F12.9))")  "f =", f
    
    write (u, "(A)")
    write (u, "(A)")  "* Recover x from momenta"
    write (u, "(A)")
    
    q = interaction_get_momenta (sf_int%interaction_t, outgoing=.true.)
    call sf_int%final ()
    deallocate (sf_int)
    
    call data%allocate_sf_int (sf_int)
    call sf_int%init (data)
    call sf_int%set_beam_index ([1])
    
    call sf_int%seed_kinematics ([k])
    call interaction_set_momenta (sf_int%interaction_t, q, outgoing=.true.)
    call sf_int%recover_x (x)
    
    write (u, "(A,9(1x,F12.9))")  "x =", x
    
    write (u, "(A)")
    write (u, "(A)")  "* Evaluate EWA structure function"
    write (u, "(A)")
    
    select type (sf_int)
    type is (ewa_t)
       sf_int%q = 1000._default
    end select
    call sf_int%complete_kinematics (x, f, r, map=.false.) 
    call sf_int%apply (scale = 100._default)
    call sf_int%write (u)
    
    call sf_int%final ()
    deallocate (sf_int)
    deallocate (x)
    deallocate (r)
    deallocate (data)

    write (u, "(A)")
    write (u, "(A)")  "* Initialize structure-function object"
    write (u, "(A)")

    allocate (ewa_data_t :: data)
    select type (data)
    type is (ewa_data_t)
       call data%init (model, flv, 0.01_default, 10._default, &
            500._default, 5000._default, .false., .true., 5._default)       
       call data%set_id (24)
    end select
    
    call data%allocate_sf_int (sf_int)
    call sf_int%init (data)
    call sf_int%set_beam_index ([1])
    
    call sf_int%write (u)

    write (u, "(A)")
    write (u, "(A)")  "* Initialize incoming momentum with E=500"
    write (u, "(A)")
    E = 500
    k = vector4_moving (E, sqrt (E**2 - flavor_get_mass (flv)**2), 3)
    call pacify (k, 1e-10_default)
    call vector4_write (k, u)
    call sf_int%seed_kinematics ([k])
        
    write (u, "(A)")
    write (u, "(A)")  "* Set kinematics for x=0.5/0.5/0.25, with EWA mapping, "
    write (u, "(A)")  "          non-coll., keeping energy"
    write (u, "(A)")
    
    allocate (r (data%get_n_par ()))
    allocate (x (size (r)))
    
    r = [0.5_default, 0.5_default, 0.25_default]
    select type (sf_int)
    type is (ewa_t)    
       sf_int%q = 1000._default
    end select
    call sf_int%complete_kinematics (x, f, r, map=.true.)
    call interaction_pacify_momenta (sf_int%interaction_t, 1e-10_default)
    call sf_int%write (u)
    
    write (u, "(A)")
    write (u, "(A,9(1x,F12.9))")  "x =", x
    write (u, "(A,9(1x,F12.9))")  "f =", f
    
    write (u, "(A)")
    write (u, "(A)")  "* Recover x and r from momenta"
    write (u, "(A)")
    
    q = interaction_get_momenta (sf_int%interaction_t, outgoing=.true.)
    call sf_int%final ()
    deallocate (sf_int)
    
    call data%allocate_sf_int (sf_int)
    call sf_int%init (data)
    call sf_int%set_beam_index ([1])
    
    call sf_int%seed_kinematics ([k])
    call interaction_set_momenta (sf_int%interaction_t, q, outgoing=.true.)
    call sf_int%recover_x (x)
    select type (sf_int)
    type is (ewa_t)
       sf_int%q = 1000._default
    end select
    call sf_int%inverse_kinematics (x, f, r, map=.true.)    
    
    write (u, "(A,9(1x,F12.9))")  "x =", x
    write (u, "(A,9(1x,F12.9))")  "r =", r
    
    write (u, "(A)")
    write (u, "(A)")  "* Evaluate EWA structure function"
    write (u, "(A)")
    
    select type (sf_int)
    type is (ewa_t)
       sf_int%q = 1000._default
    end select
    call sf_int%complete_kinematics (x, f, r, map=.true.) 
    call interaction_pacify_momenta (sf_int%interaction_t, 1e-10_default)    
    call sf_int%apply (scale = 1000._default)
    call sf_int%write (u)
    
    
    write (u, "(A)")
    write (u, "(A)")  "* Cleanup"
    
    call sf_int%final ()
    call model_list_final ()
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: sf_ewa_2"

  end subroutine sf_ewa_2


end module sf_ewa
